package cn.iocoder.yudao.module.member.api.score;

import cn.iocoder.yudao.framework.redis.helper.RedisDistributedLock;
import cn.iocoder.yudao.module.member.api.score.dto.*;
import cn.iocoder.yudao.module.member.dal.dataobject.memberUserScore.MemberUserScoreDO;
import cn.iocoder.yudao.module.member.enums.ScoreOperateTypeEnum;
import cn.iocoder.yudao.module.member.enums.ScoreSourceTypeEnum;
import cn.iocoder.yudao.module.member.service.memberUserScore.MemberUserScoreService;
import cn.iocoder.yudao.module.member.service.memberUserScoreDetail.MemberUserScoreDetailService;
import cn.iocoder.yudao.module.member.service.memberUserScoreLog.MemberUserScoreLogService;
import cn.iocoder.yudao.module.member.vo.memberUserScoreLog.MemberUserScoreLogCreateReq;
import com.alibaba.excel.util.CollectionUtils;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;

import java.util.List;
import java.util.stream.Collectors;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.infra.enums.ErrorCodeConstants.GET_LOCK_FAILED;
import static cn.iocoder.yudao.module.member.enums.ErrorCodeConstants.*;

@Slf4j
@Service
public class MemberUserScoreApiImpl implements MemberUserScoreApi{

    @Resource
    private MemberUserScoreLogService logService;
    @Resource
    private MemberUserScoreDetailService scoreDetailService;
    @Resource
    private MemberUserScoreService memberUserScoreService;
    @Resource
    private RedisDistributedLock redisDistributedLock;
    @Override
    @Transactional(rollbackFor = Exception.class)
    public MemberUserScoreOperateRespDTO operateScore(MemberUserScoreOperateReqDTO req) {
        if (req.getMemberId() == null) {
            throw exception(MEMBER_ID_IS_NULL);
        }
        if (req.getScoreCount() <= 0) {
            throw exception(SCORE_COUNT_ERROR);
        }
        String lockKey = "member:operate:socre:" + req.getMemberId();
        boolean lock = redisDistributedLock.lock(lockKey, 5000, 3, 100);
        if (!lock) {
            throw exception(GET_LOCK_FAILED);
        }
        try {
            Long logId = saveScoreLog(req);
            saveScoreDetail(req, logId);
            updateUserScore(req);
        } finally {
            redisDistributedLock.releaseLock(lockKey);
        }
        return MemberUserScoreOperateRespDTO.success(req);
    }

    @Override
    @Transactional
    public List<MemberUserScoreOperateRespDTO> batchOperateScore(MemberUserScoreBatchOperateReqDTO req) {
        if (CollectionUtils.isEmpty(req.getMemberIds())) {
            throw exception(MEMBER_ID_IS_NULL);
        }
        if (req.getScoreCount() <= 0) {
            throw exception(SCORE_COUNT_ERROR);
        }
        // 校验用户当前积分是否满足扣减要求
        if (req.getOperateType() == ScoreOperateTypeEnum.REDUCE) {
            LambdaQueryWrapper<MemberUserScoreDO> wrapper = Wrappers.lambdaQuery();
            wrapper.in(MemberUserScoreDO::getMemberId, req.getMemberIds());
            List<MemberUserScoreDO> userScoreDOList = memberUserScoreService.list(wrapper);
            List<MemberUserScoreDO> notEnoughScoreList = userScoreDOList.stream()
                    .filter(item -> item.getHoldScore() < req.getScoreCount()).collect(Collectors.toList());
            if (!notEnoughScoreList.isEmpty()) {
                throw exception(MEMBER_SCORE_NOT_ENOUGH);
            }
        }
        return req.getMemberIds().stream().map(memberId -> operateScore(MemberUserScoreOperateReqDTO.builder()
                .memberId(memberId)
                .scoreCount(req.getScoreCount())
                .operateType(req.getOperateType())
                .sourceType(req.getSourceType())
                .extParam(req.getExtParam())
                .build()))
                .collect(Collectors.toList());
    }

    private void updateUserScore(MemberUserScoreOperateReqDTO req) {
        memberUserScoreService.updateUserScore(MemberUserScoreUpdateReqDTO.builder()
                        .memberId(req.getMemberId())
                        .scoreCount(req.getScoreCount())
                        .operateType(req.getSourceType() == ScoreSourceTypeEnum.MANUAL_OPERATE ?
                                req.getOperateType() : req.getSourceType().getOperateType())
                        .sourceType(req.getSourceType())
                .build());
    }

    private void saveScoreDetail(MemberUserScoreOperateReqDTO req, Long scoreLogId) {
        scoreDetailService.updateScoreDetail(MemberUserScoreDetailUpdateReqDto.builder()
                        .memberId(req.getMemberId())
                        .scoreCount(req.getScoreCount())
                        .operateType(req.getSourceType() == ScoreSourceTypeEnum.MANUAL_OPERATE ?
                                req.getOperateType() : req.getSourceType().getOperateType())
                        .sourceType(req.getSourceType())
                        .scoreLogId(scoreLogId)
                        .expireDays(req.getExpireDays())
                .build());
    }

    private Long saveScoreLog(MemberUserScoreOperateReqDTO req) {
        return logService.createScoreLog(MemberUserScoreLogCreateReq.builder()
                .memberId(req.getMemberId())
                .scoreCount(req.getScoreCount())
                .operateType(req.getSourceType() == ScoreSourceTypeEnum.MANUAL_OPERATE ?
                        req.getOperateType().getValue() : req.getSourceType().getOperateType().getValue())
                .sourceType(req.getSourceType().getValue())
                .extParam(req.getExtParam())
                .build());
    }
}
